"""Classes and systems to handle interpolation of isochrones, given a grid of input points."""

from typing import Union, Optional, List

import numpy as np
import pandas as pd
from scipy.interpolate import LinearNDInterpolator
from scipy.interpolate import Rbf as RBFInterpolator
from scipy.interpolate import UnivariateSpline
from scipy.signal import savgol_filter
from sklearn.neighbors import NearestNeighbors


def sum_along_curve(x: Union[np.ndarray, pd.Series], y: Union[np.ndarray, pd.Series],
                    normalise: bool = True) -> Union[np.ndarray, pd.Series]:
    """Calculates differences between consecutive values along a curve and returns the cumulative sum! Assumes
    use of the Euclidean metric.

    Args:
        x (np.ndarray, pd.Series): x data for the curve.
        y (np.ndarray, pd.Series): y data for the curve.
        normalise (bool): whether or not to normalise the resulting cumulative sum.
            Default: True

    Returns:
        a np.ndarray or a pd.Series, depending on the input type.

    """

    # Calculate the difference between consecutive points
    x_difference = np.diff(x)
    y_difference = np.diff(y)

    # Calculate the Euclidean distance between each point
    distance_between_points = np.sqrt(x_difference ** 2 + y_difference ** 2)
    cumulative_distance_between_points = np.cumsum(distance_between_points)

    # Add a zero at the start because diff reduces the length by 1
    cumulative_distance_between_points = np.insert(cumulative_distance_between_points, 0, 0)

    # Normalise the distances to the range [0, 1] if desired (this makes later use a lot easier)
    if normalise:
        cumulative_distance_between_points = (cumulative_distance_between_points
                                              / np.max(cumulative_distance_between_points))

    return cumulative_distance_between_points


def find_nearest_point(points_to_match: np.ndarray, points_on_line: np.ndarray,
                       return_raw_distances: bool = False) -> Union[np.ndarray, List[np.ndarray]]:
    """Given a first and a second set of points, this function will find which second point is nearest to each first
    point, and return an array detailing these findings with the same shape as the first set of points.

    Args:
        points_to_match (np.ndarray): points to match to the line. Should have shape (n_match_points, n_features).
        points_on_line (np.ndarray): points to cross-match the points_to_match with. Should have shape
            (n_line_points, n_features).
        return_raw_distances (bool): whether to return the raw distances to each point as well or not.

    Returns:
        return_raw_distances = False:
            An array of length n_match_points where each value is the nearest point in points_on_line to that
            point_to_match.
        return_raw_distances = True:
            Same as above, but also the raw distances too (tagged on as the second item in a list).

    """
    # Raise an error if one of the arrays isn't 2D
    if len(points_to_match.shape) != 2 or len(points_on_line.shape) != 2:
        raise ValueError('Input arrays must be two dimensional.')

    # Raise an error if the last shape of the arrays isn't the same
    if points_to_match.shape[1] != points_on_line.shape[1]:
        raise ValueError('Number of features mismatch between points_to_match and points_on_line.')

    # Given that points_to_match has shape (a, n) and points_on_line has shape (b, n), we tile both arrays to have shape
    # (b, a, n) so that they can be acted on together.
    points_on_line_tiled = np.tile(points_on_line, (points_to_match.shape[0], 1, 1))
    points_to_match_tiled = (np.repeat(points_to_match, points_on_line.shape[0], axis=0)
                             .reshape(points_on_line_tiled.shape))

    # Then, we find the Euclidean distance between each point, and use the smallest one to match points to the line.
    raw_distances = points_to_match_tiled - points_on_line_tiled
    raw_distances_squared = np.sum(raw_distances**2, axis=2)
    closest_point = np.argmin(raw_distances_squared, axis=1)

    # Return raw distances to each closest point too if desired
    # Todo: is this still needed? It isn't really if I still sort
    if return_raw_distances:
        return [closest_point, raw_distances[np.arange(points_to_match.shape[0]), closest_point, :]]
    else:
        return closest_point


def proximity_to_line_sort(points_to_match: np.ndarray, points_on_line: np.ndarray) -> np.ndarray:
    """Given a field of unorganised points and an ordered list of points defining a line, this function returns the
    field of points sorted by order of how close to the points on the line they are. The final step is to sort matches
    to the same point based on their last axis distance from the point (in 2D, that will be the y axis.)

    Todo: is it ok to just oversample this, or should I improve the final sorting scoring step? This isn't an issue if
        the line has enough points, but that may often not be the case...
        Instead, it could use the derivative of the points on the line to work out which way the line is going, and then
        use that to work out if points_to_match are above or below the direction of travel. Fucking hard though!

    Notes:
        - This function was made to sort unordered CMDs.
        - Make sure the line has enough points, as the final sorting step between points matched to the same line point
            is ONLY done on y values, which will cause icky discontinuities if the line doesn't have a high enough
            resolution.

    Args:
        points_to_match (np.ndarray): points to match to the line. Should have shape (n_match_points, n_features).
        points_on_line (np.ndarray): points to cross-match the points_to_match with. Should have shape
            (n_line_points, n_features).

    Returns:
        The optimum set of arguments to sort the set of points along the line.

    """
    # Get both the nearest point to the line *and* how far away from that point it is
    closest_point, raw_distances = find_nearest_point(points_to_match, points_on_line, return_raw_distances=True)

    # Grab all the y distances from the points, and normalise them to be in the range [-0.495, 0.495]
    y_distances = raw_distances[:, -1]
    min_y = np.min(y_distances)
    max_y = np.max(y_distances)
    normalised_y_distances = ((y_distances - min_y) / (max_y - min_y) - 0.5) * 0.99

    # Add the y distances to the closest point ordering, and then sort the points given that we're sorting on both point
    # along the line (integer steps) and how above/below the point each one is (float steps)
    sort_args = np.argsort(closest_point + normalised_y_distances)

    return sort_args


def nn_graph_sort(x, y):
    """Sorts data based on a nearest neighbour graph.

    # Todo: still needed?

    Note:
        - Currently always starts at the lowest value.
        - A solution to my CMDInterpolator sorting problem from:
            https://stackoverflow.com/questions/37742358/sorting-points-to-form-a-continuous-line
        - MAY NOT RETURN A GRAPH INCLUDING ALL POINTS - this solution is quite basic unfortunately and does NOT work for
            noisy data.

    Args:
        x (np.ndarray): x of points to sort.
        y (np.ndarray): y of points to sort.

    Returns:
        The optimum set of arguments to sort the path.


    """
    import networkx as nx

    # Order stuff into a 2D array
    points = np.column_stack([x, y])

    # Create a graph (as a sparse matrix), where each node is connected to its nearest neighbours:
    clf = NearestNeighbors(2).fit(points)
    graph = clf.kneighbors_graph()
    nx_graph = nx.from_scipy_sparse_matrix(graph)

    # Extract the best paths between all the points
    # Todo optimise this steaming pile of for-loop dogshit!!!1!

    list_of_all_paths = [list(nx.dfs_preorder_nodes(nx_graph, source=i)) for i in range(x.size)]

    # Cycle over all paths and find the one with the smallest cost
    minimum_distance = np.inf
    path_with_minimum_distance = None

    for i in range(len(points)):
        a_path = list_of_all_paths[i]  # order of nodes
        ordered = points[a_path]  # ordered nodes
        # Find cost of that order by the sum of euclidean distances between points (i) and (i+1),
        cost = (((ordered[:-1] - ordered[1:]) ** 2).sum(1)).sum()
        if cost < minimum_distance:
            minimum_distance = cost
            path_with_minimum_distance = i

    return list_of_all_paths[path_with_minimum_distance]


class IsochroneInterpolator:
    def __init__(self, data_isochrone: pd.DataFrame, parameters_as_arguments: Union[list, tuple] = ('MH', 'logAge'),
                 parameters_to_infer: Union[list, tuple] = ('G_BP-RP', 'Gmag'), interpolation_type: str = 'LinearND'):
        """Given an input in the format outputted by the CMD 3.3 web interface, this class will produce interpolated
        isochrones, marginalised over the specified argument_parameters. Yay!

        Todo: consider removing RadialBasisFunction interpolation, as it can't use duplicated co-ordinates of any kind
            which is... pretty crappy.
            See: https://stackoverflow.com/questions/27295853/rbf-interpolation-fails-linalgerror-singular-matrix

        Args:
            data_isochrone (pd.DataFrame): the data for the isochrones, in the same format as CMD 3.3 (so: a long
                fuckoff list of isochrones, all joined together.)
            parameters_as_arguments (list, optional): parameters to use as arguments in the interpolator (e.g. logAge
                for each isochrone).
                Default: None, which uses ['logZini', 'logAge'] as default.
            parameters_to_infer (str, optional): parameters to infer with the interpolator. There must be two due to the
                design of this implementation!
                Default: ('G_BP-RP', 'Gmag')
            interpolation_type (str): type of interpolation to use.
                Implemented: 'LinearND' : scipy.interpolate.LinearNDInterpolator
                             'RadialBasisFunction' : scipy.interpolate.Rbf
                Default: 'LinearND'

        Returns:
            absolutely... NOTHING
            (call the class if you want to evaluate it)
        """
        # Set all the parameter names to the class
        self.argument_parameters = parameters_as_arguments
        self.inferred_parameters = parameters_to_infer

        # Calculate the normalised distances along the curve
        data_isochrone = self.calculate_normalised_cumulative_distances_along_curve(data_isochrone)

        # Hence, grab the input data we'll need for interpolation
        input_data = data_isochrone[self.argument_parameters + ['cumulative_distance_along_curve']].to_numpy()

        # Interpolate each of the inferred_parameters as a function of the normalised distance parameter.
        if interpolation_type == 'LinearND':
            # Interpolate!!!
            self.x_interpolator = LinearNDInterpolator(input_data, data_isochrone[self.inferred_parameters[0]])
            self.y_interpolator = LinearNDInterpolator(input_data, data_isochrone[self.inferred_parameters[1]])

            # Set __call__ to use this function!
            self.evaluate = self.evaluate_linear_nd_interpolator

        elif interpolation_type == 'RadialBasisFunction':
            # Prepare the input for use with the * way of unpacking numpy axes
            input_data = input_data.T
            input_data_x = np.vstack((input_data, data_isochrone[self.inferred_parameters[0]].to_numpy()))
            input_data_y = np.vstack((input_data, data_isochrone[self.inferred_parameters[1]].to_numpy()))

            # Interpolate!!!
            self.x_interpolator = RBFInterpolator(*input_data_x)
            self.y_interpolator = RBFInterpolator(*input_data_y)

            # Set __call__ to use this function!
            self.evaluate = self.evaluate_rbf_interpolator

        else:
            raise ValueError('Specified interpolation method not recognised!')

    def calculate_normalised_cumulative_distances_along_curve(self, data_isochrone: pd.DataFrame):
        """Calculates cumulative distances along an arbitrary curve and normalises them.

        Args:
            data_isochrone (pd.DataFrame): the data describing the isochrone(s) in the same output format as CMD 3.3.

        Returns:
            the modified data_isochrone, with cumulative distances under the key 'cumulative_distance_along_curve'.
        """

        # Grab the number of argument_parameters (this comment is self-evident)
        n_parameters = len(self.argument_parameters)

        if n_parameters != 0:

            # Initialise the new column as nans, so that an error will be thrown if any are missed
            data_isochrone['cumulative_distance_along_curve'] = np.nan

            # Grab all possible unique parameter combinations
            unique_parameter_combos = np.unique(data_isochrone[self.argument_parameters], axis=0)

            # Cycle over all possible parameter combinations
            for a_parameter_combination in unique_parameter_combos:
                # Sum the number of parameter matches per row
                stars_to_use = np.sum(data_isochrone[self.argument_parameters] == a_parameter_combination, axis=1)

                # Only accept stars where all the parameters_are matched (the row-wise sum should equal n_parameters)
                stars_to_use = np.asarray(stars_to_use == n_parameters)

                # Hence, calculate the distance
                x = data_isochrone.loc[stars_to_use, self.inferred_parameters[0]]
                y = data_isochrone.loc[stars_to_use, self.inferred_parameters[1]]
                data_isochrone.loc[stars_to_use, 'cumulative_distance_along_curve'] = sum_along_curve(x, y)

        # If there aren't any other parameters, then we just find the cumulative distance for the whole damn table
        else:
            x = data_isochrone[self.inferred_parameters[0]]
            y = data_isochrone[self.inferred_parameters[1]]
            data_isochrone['cumulative_distance_along_curve'] = sum_along_curve(x, y)

        return data_isochrone

    def evaluate_linear_nd_interpolator(self, input_data: np.ndarray) -> list:
        """Evaluation function in the case that the class's interpolator is a scipy.interpolate.LinearNDInterpolator.

        Args:
            input_data (np.ndarray): input in shape (n_points, n_dims)

        Returns:
            a list containing [x_data, y_data] for the requested points.
        """
        return [self.x_interpolator(input_data), self.y_interpolator(input_data)]

    def evaluate_rbf_interpolator(self, input_data: np.ndarray) -> list:
        """Evaluation function in the case that the class's interpolator is a scipy.interpolate.Rbf (radial basis
        function interpolator.)

        Args:
            input_data (np.ndarray): input in shape (n_points, n_dims)

        Returns:
            a list containing [x_data, y_data] for the requested points.
        """
        input_data = input_data.T
        return [self.x_interpolator(*input_data), self.y_interpolator(*input_data)]

    def __call__(self, input_data: Union[np.ndarray, pd.DataFrame],
                 range_to_evaluate: Union[str, np.ndarray] = 'full isochrone',
                 resolution: int = 100) -> list:
        """Evaluate the interpolator against a specific set of points. Contains systems to automate a lot of the process
        for you, because it's kind like that. =)

        Args:
            input_data (np.ndarray, pd.DataFrame): input arguments for isochrones to be evaluated at, with shape
                (n_points, n_dims). Every n_points will generate a full isochrone.
            range_to_evaluate (str, np.ndarray): array of values in the correct range (typically [0, 1]) to evaluate the
                isochrone at.
                Default: 'full isochrone', which computes a set of points in the range [0, 1] automatically.
            resolution (int): if range_to_evaluate is set to 'full isochrone', this controls how many points will be
                returned.

        Returns:
            a list containing [x_data, y_data] for the requested points.
        """
        if range_to_evaluate == 'full isochrone':
            range_to_evaluate = np.linspace(0, 1, resolution)

        points_per_isochrone = range_to_evaluate.size

        # Tile the range_to_evaluate to match the length of input_data's number of points
        range_to_evaluate = np.tile(range_to_evaluate, [input_data.shape[0]])

        # Process the function arguments into something feedable to the interpolator
        input_data = np.repeat(np.asarray(input_data), points_per_isochrone, axis=0)
        input_data = np.hstack((input_data, np.expand_dims(range_to_evaluate, axis=1)))

        return self.evaluate(input_data)


class CMDInterpolator:
    def __init__(self,
                 parameters: Union[list, tuple] = ('B-R', 'G'),
                 curve_parameterisation_type: str = 'summed',
                 data_sorted_on: str = 'y_values',
                 filter_input: bool = False,
                 filter_window_size: int = 5,
                 filter_order: int = 3,
                 interp_type: str = 'UnivariateSpline',
                 interp_weights: Optional[str] = None,
                 interp_smoothing: Optional[float] = 0.05,
                 interp_order: int = 3):
        """Interpolates data from a potential cluster, allowing its scatter from the best fit spline to be evaluated.

        Notes:
            - Call the object later to evaluate the interpolated spline!
            - Will get slow if an especially big cluster is passed to the method.

        Args:
            parameters (list, tuple): the parameters in data_gaia that define the blue-red and total magnitudes.
            curve_parameterisation_type (string): how to parameterise the curve. Options:
                linear: parameterise with a sequence from 0 to 1 for all data points. Easiest to do,
                    but may produce sharp changes in gradient in what has to be fit.
                summed:  parameterise by mapping the sequence from 0 to 1 to a weighted sum of the distance along the
                    curve. Will generally produce better results, but assumes that the sorting was largely successful.
                Default: 'summed'
            data_sorted_on (string): how to sort the data (since the interpolator has to receive points in a certain order.
                y_values: sort all data on the magnitude parameter.
                x_values: sort all data on the colour parameter.
                Default: 'y_values'
            Todo add filtering to the docstr
            interp_type (string): type of interpolator to use. Currently, only 'UnivariateSpline' is implemented.
            interp_weights (string, optional): name of the column in data_gaia that gives parameter weights. If None, no
                weights are passed to the spline fit.
                Default: None
            interp_smoothing (float, optional): amount of smoothing to apply. If None, the SciPy intepolator will
                try to calculate an optimum value itself.
                Default: 0.05
            interp_order (int): order of the polynomial to fit.
                Default: 3
        """
        # Define some useful stuff to keep class-side
        self.fits_completed = 0

        # Set the parameter names to the class
        self.parameters = parameters
        self.weight_parameter = interp_weights

        # Write Useful stuff to the class
        self.sort_method = data_sorted_on
        self.curve_parameterisation_type = curve_parameterisation_type
        self.curve_parameter = None
        self.filter_input = filter_input
        self.filter_window_size = filter_window_size
        self.filter_order = filter_order

        # Stuff we'll need to set later to do with interpolation
        self.x_interpolator = None
        self.y_interpolator = None
        self.interpolation_type = interp_type

        if self.interpolation_type == 'UnivariateSpline':
            self.interpolation_arguments = {'s': interp_smoothing, 'k': interp_order}

            # Set __call__ to use this function!
            self.evaluate = self.evaluate_univariate_spline

        else:
            raise ValueError('Specified interpolation method not recognised!')

    def fit(self, data_gaia, max_repeats: int = 1, print_current_step: bool = False):
        """Fits a spline to the CMD data. Depending on the sort method chosen, this may be done iteratively to improve
        the quality of the fit.

        Args:
            data_gaia (pd.DataFrame): the Gaia data, which ought to contain magnitude information (including a blue -
                red column.)
            max_repeats (int): how many times to re-run the fitting.


        """
        # Kill The Panda, It's Time For X and Y Arrays
        data_colour = np.asarray(data_gaia[self.parameters[0]])
        data_magnitude = np.asarray(data_gaia[self.parameters[1]])

        # Decide on what the weights for the curve fitting will be
        if self.weight_parameter is not None:
            star_weights = data_gaia[self.weight_parameter]
        else:
            star_weights = None  # Aka no weights will be passed to functions

        while self.fits_completed < max_repeats:
            # Sort the data
            data_colour, data_magnitude = self._input_sort(data_colour, data_magnitude)

            # Make a parameter to describe the 2D curve with
            self.curve_parameter = self._input_parameterisation(data_colour, data_magnitude)

            # Filter the input (assuming that filtering was turned on during __init__)
            filtered_data_colour, filtered_data_magnitude = self._input_filter(data_colour, data_magnitude)

            # Interpolate!
            self._input_interpolate(filtered_data_colour, filtered_data_magnitude, star_weights)

            self.fits_completed += 1

            if print_current_step:
                print(self.fits_completed)

    def _input_sort(self, data_colour, data_magnitude):
        """Sorts input values to the class."""

        # Decide on how we're gonna initially sort the data
        # Sort x and y based on y values
        if self.sort_method == 'y_values':
            sort_args = np.argsort(data_magnitude)
            self.fits_completed = np.inf  # This sort method is non-iterable!

        # Sort x and y based on x values
        elif self.sort_method == 'x_values':
            sort_args = np.argsort(data_colour)
            self.fits_completed = np.inf  # This sort method is non-iterable!

        # Create a nearest neighbour graph and sort using it
        elif self.sort_method == 'nearest_neighbour_graph':
            sort_args = nn_graph_sort(data_colour, data_magnitude)
            self.fits_completed = np.inf  # This sort method is non-iterable!

        # Sort based on proximity to a pre-established line
        elif self.sort_method == 'proximity_to_line':
            # If this is our first run, then we have to sort on y values first instead
            if self.fits_completed == 0:
                sort_args = np.argsort(data_magnitude)
            else:
                # Grab data to feed to the sorter
                points_to_match = np.vstack([data_colour, data_magnitude]).T

                # Use a resoltuion of 100 or twice the length of the array & grab interpolated points
                resolution = np.max([1 * data_colour.size, 100])
                points_on_line = np.vstack(self(resolution=resolution)).T

                sort_args = proximity_to_line_sort(points_to_match, points_on_line)
        else:
            raise ValueError("Selected sort_method not recognised!")

        data_colour = data_colour[sort_args]
        data_magnitude = data_magnitude[sort_args]

        return data_colour, data_magnitude

    def _input_parameterisation(self, data_colour, data_magnitude):
        """Parameterise the CMD."""
        # Decide on how we're gonna parameterise the curve
        if self.curve_parameterisation_type == 'linear':
            curve_parameter = np.linspace(0, 1, data_magnitude.size)
        elif self.curve_parameterisation_type == 'summed':
            curve_parameter = sum_along_curve(data_colour, data_magnitude, normalise=True)
        else:
            raise ValueError('Specified method for curve_parameterisation_type of the curve not recognised!')

        return curve_parameter

    def _input_filter(self, data_colour, data_magnitude):
        """Filters the input data."""

        if self.filter_input:

            # # Adapt the filtering strength based on how many fits we've already made
            # self.filter_window_fraction /= self.fits_completed + 1

            # Set the filter window length to be some fraction of the dataset size, making sure that it's an odd number
            # filter_window_length = int(data_colour.size * self.filter_window_fraction)
            #
            # # Make sure it isn't less than the polyorder
            # if filter_window_length <= self.filter_order:
            #     filter_window_length = self.filter_order + 1
            #
            # # Make sure it's odd
            # filter_window_length += (1 - filter_window_length % 2)
            # filter_window_size = filter_window_length

            # Apply the filter!
            data_colour = savgol_filter(data_colour, self.filter_window_size, self.filter_order)
            data_magnitude = savgol_filter(data_magnitude, self.filter_window_size, self.filter_order)

        return data_colour, data_magnitude

    def _input_interpolate(self, data_colour, data_magnitude, star_weights):
        """Interpolates, given settings in the class"""
        if self.interpolation_type == 'UnivariateSpline':
            self.x_interpolator = UnivariateSpline(self.curve_parameter, data_colour, w=star_weights,
                                                   **self.interpolation_arguments)
            self.y_interpolator = UnivariateSpline(self.curve_parameter, data_magnitude, w=star_weights,
                                                   **self.interpolation_arguments)

        else:
            raise ValueError('Specified interpolation method not recognised!')

        return data_colour, data_magnitude

    def evaluate_univariate_spline(self, input_data: np.ndarray) -> tuple:
        """Evaluation function in the case that the class's interpolator is a scipy.interpolate.UnivariateSpline.

        Args:
            input_data (np.ndarray): input parameter data at whatever resolution is desired.

        Returns:
            a list containing [x_data, y_data] for the requested points.
        """
        return self.x_interpolator(input_data), self.y_interpolator(input_data)

    def __call__(self, range_to_evaluate: Union[str, np.ndarray] = 'full isochrone',
                 resolution: int = 100) -> tuple:
        """Evaluate the interpolator and re-produce all (or part of) the interpolated isochrone. Contains systems to
        automate a lot of the process for you, because it's kind like that. =)

        Args:
            range_to_evaluate (str, np.ndarray): array of values in the correct range (typically [0, 1]) to evaluate the
                isochrone at.
                Default: 'full isochrone', which computes a set of points in the range [0, 1] automatically.
            resolution (int): if range_to_evaluate is set to 'full isochrone', this controls how many points will be
                returned.

        Returns:
            a list containing [x_data, y_data] for the requested points.
        """
        if range_to_evaluate == 'full isochrone':
            range_to_evaluate = np.linspace(0, 1, resolution)

        return self.evaluate(range_to_evaluate)


class IsochroneDistanceInterpolator:
    def __init__(self):
        raise NotImplementedError("What are you doing, this isn't implemented, stop")

    def __call__(self):
        pass




"""A set of tests for use with the pytest module, covering ocelot.plot"""

# FUCKING HATE PYTHON IMPORTS AAAA
# (the below fixes this though)
try:
    from .context import ocelot
except ModuleNotFoundError:
    print('Unable to find ocelot via .context! Trying to import from your python path instead...')
try:
    import ocelot
except ModuleNotFoundError:
    raise ModuleNotFoundError('Unable to find ocelot')

import pickle
import pandas as pd
from pathlib import Path

import numpy as np
import pytest

# Path towards the test isochrones
max_label = 0
path_to_test_isochrones = Path('./test_data/isochrones/isochrones.dat')
#path_to_test_isochrone = Path('../../../data/isochrones/191015_isochrones_z-2_8_to_1_2/z_-2.8_1.2_1young_age.dat')

# Path towards blanco 1 data and clustering produced by DBSCAN
path_to_blanco_1 = Path('./test_data/blanco_1_gaia_dr2_gmag_18_cut.pickle')
path_to_blanco_1_cluster_labels = Path('./test_data/blanco_1_gaia_dr2_gmag_18_cut_LABELS.pickle')


def test_curve_normalisation():
    """Tests the curve normalisation function ocelot.plot.utilities.normalise_a_curve."""
    # Make a shitty little sine wave
    x_range = np.linspace(0, 10, num=100)
    y_range = np.sin(x_range)

    # Calculate some areas
    true_area = np.trapz(y_range, x=x_range)
    normalised_area = np.trapz(ocelot.plot.utilities.normalise_a_curve(x_range, y_range, 2.5), x=x_range)
    unnormalised_area = np.trapz(ocelot.plot.utilities.normalise_a_curve(x_range, y_range, 0.0), x=x_range)

    # Test that normalisation works to an arbitrary number
    assert np.allclose(2.5, normalised_area, rtol=0.0, atol=1e-8)

    # Test that normalisation can be turned off
    assert np.allclose(true_area, unnormalised_area, rtol=0.0, atol=1e-8)


def test_percentile_based_plot_limits():
    """Tests the function ocelot.plot.utilities.percentile_based_plot_limits, also testing the function
    ocelot.plot.utilities.good_points_plot_limits at the same time."""
    # Make some test data
    x = np.linspace(0, 100, num=101)
    y = x**2

    # -------------
    # Set some limits based on x, also test the padding
    x_percentiles = [0, 30]
    target_x = np.asarray([-1.5, 31.5])
    target_y = np.asarray([-45., 945.])
    x_limits, y_limits = ocelot.plot.utilities.percentile_based_plot_limits(
        x, y, x_percentiles=x_percentiles, range_padding=0.05)

    # Check x and y
    assert np.allclose(x_limits, target_x, rtol=0.0, atol=1e-8)
    assert np.allclose(y_limits, target_y, rtol=0.0, atol=1e-8)

    # -------------
    # Set some limits based on y
    y_percentiles = [50, 100]
    target_x = np.asarray([50., 100.])
    target_y = np.asarray([2500., 10000.])
    x_limits, y_limits = ocelot.plot.utilities.percentile_based_plot_limits(
        x, y, y_percentiles=y_percentiles, range_padding=None)

    # Check x and y
    assert np.allclose(x_limits, target_x, rtol=0.0, atol=1e-8)
    assert np.allclose(y_limits, target_y, rtol=0.0, atol=1e-8)

    # -------------
    # Set some limits based on x and y at the same time
    x_percentiles = [0, 50]
    y_percentiles = [0, 10]
    target_x = np.asarray([0., 50.])
    target_y = np.asarray([0., 100.])
    x_limits, y_limits = ocelot.plot.utilities.percentile_based_plot_limits(
        x, y, x_percentiles=x_percentiles, y_percentiles=y_percentiles, range_padding=None)

    # Check x and y
    assert np.allclose(x_limits, target_x, rtol=0.0, atol=1e-8)
    assert np.allclose(y_limits, target_y, rtol=0.0, atol=1e-8)

    # -------------
    # Check that specifying nothing raises a value error
    with pytest.raises(ValueError, match="One or both of x_percentiles and y_percentiles must be specified."):
        ocelot.plot.utilities.percentile_based_plot_limits(x, y)


def test_isochrone_plotting(show_figure=False):
    """Integration test for the function ocelot.plot.isochrone."""
    my_isochrones = ocelot.isochrone.read_cmd_isochrone(path_to_test_isochrones, max_label=max_label)

    # Cut some stars for speed purposes
    stars_to_cut = np.asarray(my_isochrones['MH'] != 0.0).nonzero()[0]
    my_isochrones = my_isochrones.drop(stars_to_cut).reset_index(drop=True)

    # Define constants
    test_points_literature = np.asarray([[6.0], [6.5]])
    test_points_interpolated = np.asarray(np.expand_dims(np.linspace(6.0, 6.50, 4), axis=1))
    arguments = ['logAge']

    # Interpolate, so that we have something to use
    isochrones_for_fun_and_profit = ocelot.isochrone.IsochroneInterpolator(my_isochrones,
                                                                           parameters_as_arguments=['logAge'],
                                                                           interpolation_type='LinearND')

    # Check that the plotting code runs
    ocelot.plot.isochrone(data_isochrone=my_isochrones,
                          literature_isochrones_to_plot=test_points_literature,
                          isochrone_interpolator=isochrones_for_fun_and_profit,
                          interpolated_isochrones_to_plot=test_points_interpolated,
                          isochrone_arguments=arguments,
                          literature_colors='same',
                          interpolated_colors='autumn',
                          show_figure=show_figure,
                          figure_size=[10, 10])

    return my_isochrones


def test_clustering_result_plotting(show_figure=False):
    """Integration test for the function ocelot.plot.clustering_result."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    with open(path_to_blanco_1_cluster_labels, 'rb') as handle:
        labels = pickle.load(handle)

    # Grab the unique labels
    unique_labels = np.unique(labels)  # The first value is -1, which is noise points

    cluster_shading = (1 - np.clip(1. * np.sqrt((data_gaia['ra'] - np.mean(data_gaia['ra']))**2
                                                 + (data_gaia['dec'] - np.mean(data_gaia['dec']))**2),
                                   0.0, 1.0))

    # Plot time!
    fig, ax = ocelot.plot.clustering_result(
        data_gaia, labels, unique_labels[1:], cluster_shading, show_figure=show_figure,
        figure_title="Blanco 1 should be clearly visible! \nCut at G=18, clustered by DBSCAN",
        pmra_plot_limits=[14, 24], pmdec_plot_limits=[-2, 8], cmd_plot_y_limits=[6, 18],
        cluster_marker_radius=(1., 3., 1.,))

    # Friendship ended with excessive memory use,,, now "del" is my friend
    del data_gaia
    del labels

    return fig, ax


def test_location_plotting(show_figure=False):
    """Integration test for the function ocelot.plot.location."""
    # Read in data for Blanco 1
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Plot time!
    fig, ax = ocelot.plot.location(
        data_gaia, show_figure=show_figure, figure_title="Blanco 1 should be clearly visible! \nCut at G=18",
        pmra_plot_limits=[14, 24], pmdec_plot_limits=[-2, 8])

    # Friendship ended with excessive memory use,,, now "del" is my friend
    del data_gaia

    return fig, ax


def test_nearest_neighbour_plotting(show_figure=True):
    """Tests the functionality of ocelot.plot.nearest_neighbor_distances."""
    # Create some bullshit data that looks about right with a beta distribution
    distances = (np.sort(np.random.beta(4, 4, size=1000)) + 0.1).reshape(-1, 1)

    # Also make a thing to fit this with using an analytical beta distribution
    x_range = distances[::10, 0]
    y_range = np.linspace(0, 999, num=x_range.shape[0]) + 1

    fitting_func = {'x': x_range,
                    'y': y_range,
                    'style': 'r-',
                    'label': 'fit',
                    'differentiate': True}

    line = {'x': np.asarray([0.5, 0.5]),
            'y': np.asarray([0.00001, 1000]),
            'style': 'b--',
            'label': 'nn = 0.5',
            'differentiate': False}

    normalisation_constants = [1., 1., 0]

    ocelot.plot.nearest_neighbor_distances(distances, number_of_derivatives=2, figure_size=[6, 8],
                                           show_figure=show_figure, figure_title="messing with a beta function",
                                           normalisation_constants=normalisation_constants,
                                           functions_to_overplot=[fitting_func, line],
                                           show_numerical_derivatives=False)


def test_gaia_explorer():
    """Tests the Gaia explorer figure thing of fun"""
    # Firstly, we're gonna need a Blanco 1 to test with
    with open(path_to_blanco_1, 'rb') as handle:
        data_gaia = pickle.load(handle)

    # Let's guess at where the cluster is
    cluster_location = dict(name="Blanco 1", ra=1.0, dec=-30.0, pmra=18.7, pmdec=2.9, parallax=4.25,
                            ang_radius_t=2, pmra_error=1., pmdec_error=1., parallax_error=0.25)

    # Let's also make a cheeky little dataframe of fake locations to test things
    catalogue = pd.DataFrame(
        dict(name=["Blanco 1", "not Blanco", "ocelots r cute", "emily", "lauren"],
             ra=[1.0, 1.5, 1.7, 0.5, 0.0],
             dec=[-30.0, -30.5, -29.5, -30.5, -30],
             pmra=[18.7, 22, 18, 16, 0],
             pmdec=[2.9, 5, 7, -3, 0],
             parallax=[4.25, 2, 3, 1, 0],
             source=["TCG+20", "bob", "bob", "me", "me"]))

    # And plot it!
    # ocelot.plot.ion()
    gaia_explorer = ocelot.plot.GaiaExplorer(data_gaia, cluster_location, debug=True, extra_catalogue_to_plot=catalogue,
                                             extra_catalogue_cmap={'TCG+20': 'b', 'bob': 'c', 'me': 'm'},
                                             error_regions_multiplier=5.)
    gaia_explorer()


# Run tests manually if the file is ran
if __name__ == '__main__':
    #test_curve_normalisation()
    #test_percentile_based_plot_limits()
    #iso = test_isochrone_plotting(show_figure=True)
    #clustering_result = test_clustering_result_plotting(show_figure=True)
    #location = test_location_plotting(show_figure=True)
    #test_nearest_neighbour_plotting(True)
    test_gaia_explorer()



"""Some little friends to help with various odd jobs in OCELOT's plotting submod."""

from typing import Union, Optional
from scipy.stats import iqr

import numpy as np


def calculate_alpha(fig, ax, n_points: int, marker_size: float, desired_max_density: float = 0.1,
                    scale_with_dpi: bool = False,
                    max_alpha: float = 0.3,
                    min_alpha: float = 0.001,
                    scatter_plot: bool = False):
    """Calculates the optimum alpha value for points to make the plot not be
    over-saturated. Relies on the area value being correctly calculated
    prior to input, which should be the value in inches for the figure. It also
    assumes that all points are uniformly distributed across the figure area
    (which they often are not), so the input area number may need tweaking.

    Args:
        fig (matplotlib figure): the figure element. Required to calculate alpha values precisely.
        ax (matplotlib axis): the figure element. Required to calculate alpha values precisely.
        n_points (int): the number of points to plot.
        marker_size (px): the radius of the marker. Note that plt.scatter()
            specifies marker *area* but plt.plot() uses marker size, which is
            analogous to the marker radius. In the former case, ensure scatter_plot is True.
        scale_with_dpi (bool): whether or not to scale alpha values with the dpi of the figure.
            Default: False (correct for scatter plots)
        desired_max_density (float): tweakable parameter that was found to
            produce the best results. Simply mutliplying the area value is
            probably easier for normal function use, though.
            Default: 0.1.
        max_alpha (float): max alpha value to allow.
            Default: 0.3
        min_alpha (float): min alpha value to allow.
            Default: 0.001
        scatter_plot (str): whether to calculate alpha values for a scatter plot, which specifies the marker_size as an
            area, *not* a radius!

    Returns:
        float: A recommended alpha value to use on these axes.

    Todo: could still consider if the data are clumped or not

    """
    # We calculate the area of the figure with some fancy magic
    # See https://stackoverflow.com/questions/19306510/determine-matplotlib-axis-size-in-pixels
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    area = bbox.width * bbox.height

    # Scatter plots are awkward (see above function docs)
    if scatter_plot:
        marker_area = n_points * marker_size  # In units of px
    else:
        marker_area = n_points * np.pi * marker_size ** 2  # In units of px

    # We aim to have a maximum density of uniform_max_density, assuming that
    # n_points are distributed evenly across the area.
    if scale_with_dpi:
        total_area = area * fig.dpi ** 2
    else:
        total_area = area * 100**2  # If not, we just use the default area
    current_max_density = marker_area / total_area

    # We only do the alpha calculation if current max density is non-zero, just to avoid a lil error
    if current_max_density != 0:
        return np.clip(desired_max_density / current_max_density, min_alpha, max_alpha)
    else:
        return max_alpha


def normalise_a_curve(x_values: np.ndarray, y_values: np.ndarray, normalisation_constant: Union[int, float]):
    """Wrapper for np.trapz intended for use with ocelot.plot.axis.nn_statistics.point_number_vs_nn_distance.

    Args:
        x_values (np.ndarray): x values of the curve.
        y_values (np.ndarray): y values of the curve.
        normalisation_constant (float): area to set the curve to have. If zero, the curve will not be normalised.

    Returns:
        normalised y_values (or unmodified y_values if normalisation_constant is 0)

    """
    if normalisation_constant == 0:
        return y_values
    else:
        return y_values * normalisation_constant / np.trapz(y_values, x=x_values)


def _good_points_plot_limits(target_points: np.ndarray, constraint_points: np.ndarray, constraint_limits: np.ndarray):
    """Finds appropriate plot limits for a target array of points based on the limits of a constraint set of points.
    Solves the issue with matplotlib limits being set on one dimension but not cropping out blank space on the other.

    Args:
        target_points (np.ndarray): points to find limits for
        constraint_points (np.ndarray): the points that already have a cut-off constraint
        constraint_limits (np.ndarray): the limits for the pre-constrained points.

    Returns:
        target_point_limits (np.ndarray): limits for the target points that will remove blank space.

    """
    good_points = np.logical_and(constraint_points >= constraint_limits[0], constraint_points <= constraint_limits[1])
    return np.asarray([target_points[good_points].min(), target_points[good_points].max()])


def percentile_based_plot_limits(x: np.ndarray,
                                 y: np.ndarray,
                                 x_percentiles: Optional[Union[tuple, list, np.ndarray]] = None,
                                 y_percentiles: Optional[Union[tuple, list, np.ndarray]] = None,
                                 range_padding: Optional[float] = 0.05):
    """Derives appropriate limits for a 2D plot based on percentiles of the data. May use just one or both of x or y
    data to infer plot limits. When only one is used, the unused axis will be cropped to remove blank space (which
    solves the issue with matplotlib limits being set on one dimension but not cropping out blank space on the other.)

    Args:
        x (np.ndarray): x points to base plotting limits from.
        y (np.ndarray): y points to base plotting limits from.
        x_percentiles (list-like): length two array of x percentiles to plot. Values within must be in the range
            [0, 100]. If None, x plot limits are inferred from the x range of valid y points.
        y_percentiles (np.ndarray, list): length two array of y percentiles to plot. Values within must be in the range
            [0, 100]. If None, y plot limits are inferred from the y range of valid x points.
        range_padding (list-like): padding factor to apply based on the valid range of the data. Stops plots from
            being cropped exactly on the points at hand. Remember padding is added individually around all ends of the
            data! May be set to "None" to be turned off.
            Default: 0.05

    """
    # Make sure the inputs are flat numpy arrays
    x = np.asarray(x).flatten()
    y = np.asarray(y).flatten()

    # First, we make some actual limits
    if x_percentiles is not None:
        x_plot_limits = np.percentile(x, x_percentiles)

        # When x and y specified
        if y_percentiles is not None:
            y_plot_limits = np.percentile(y, y_percentiles)

        # When only x specified
        else:
            y_plot_limits = _good_points_plot_limits(y, x, x_plot_limits)

    # When only y specified
    elif y_percentiles is not None:
        y_plot_limits = np.percentile(y, y_percentiles)
        x_plot_limits = _good_points_plot_limits(x, y, y_plot_limits)

    # When neither specified: value error!
    else:
        raise ValueError("One or both of x_percentiles and y_percentiles must be specified.")

    # We can also add some padding if requested
    if range_padding is not None:
        # Grab the stars within the ranges
        good_x = np.logical_and(x >= x_plot_limits[0], x <= x_plot_limits[1])
        good_y = np.logical_and(y >= y_plot_limits[0], y <= y_plot_limits[1])

        # Find the statistical range/peak to peak (ptp) value of the good stars
        x_padding = np.ptp(x[good_x]) * range_padding
        y_padding = np.ptp(y[good_y]) * range_padding

        # Apply the padding to the limits
        x_plot_limits += np.asarray([-x_padding, x_padding])
        y_plot_limits += np.asarray([-y_padding, y_padding])

    return x_plot_limits, y_plot_limits
